from baangt.base.utils import utils
from baangt.base.TestRunUtils import TestRunUtils
import baangt.base.GlobalConstants as GC
import baangt.base.CustGlobalConstants as CGC
import xlrd
import logging

logger = logging.getLogger("pyC")


class TestRunExcelImporter:
    def __init__(self, FileNameAndPath, testRunUtils: TestRunUtils):
        # self.testRunAttributes = testRunUtils.testRunAttributes
        self.testRunUtils = testRunUtils
        try:
            self.excelFile = xlrd.open_workbook(FileNameAndPath)
        except FileNotFoundError as e:
            raise BaseException(f"File not found - exiting {e}")

        self.fileName = utils.extractFileNameFromFullPath(FileNameAndPath)

    def importConfig(self):
        self._initTestRun()
        return self.testRunUtils.getCompleteTestRunAttributes(self.fileName)

    def _initTestRun(self):
        self.testRunUtils.testRunAttributes = \
            {
                self.fileName: {
                    GC.KWARGS_TESTRUNATTRIBUTES: {
                    },
                },
            }
        testRunAttributes = self.testRunUtils.testRunAttributes[self.fileName][GC.KWARGS_TESTRUNATTRIBUTES]

        xlsTab = self._getTab("TestRun")
        if xlsTab:
            testRunAttributes[GC.EXPORT_FORMAT] = \
                {GC.EXPORT_FORMAT: self._getValueFromList(xlsTab=xlsTab,
                                                          searchField=GC.EXPORT_FORMAT)}
        else:
            # Default Value
            testRunAttributes[GC.EXPORT_FORMAT] = {GC.EXPORT_FORMAT: GC.EXP_XLSX}

        xlsTab = self._getTab("ExportFieldList")
        if xlsTab:
            testRunAttributes[GC.EXPORT_FORMAT][GC.EXP_FIELDLIST] = self._getRowAsList(xlsTab)
        else:
            # Default Value
            testRunAttributes[GC.EXPORT_FORMAT][GC.EXP_FIELDLIST] = [GC.TESTCASESTATUS,
                                                                     GC.TIMING_DURATION,
                                                                     GC.TIMELOG]

        testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE] = {}
        testrunSequence = testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE]

        xlsTab = self._getTab("TestCaseSequence")
        if xlsTab:
            lSequenceDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            lSequenceDict = {1:
                {
                "SequenceClass": GC.CLASSES_TESTCASESEQUENCE,
                "TestDataFile": "",
                "Sheetname": "",
                "ParallelRuns": 1,
                "FromLine": 0,
                "ToLine": 999999
                },
            }
        for key, sequence in lSequenceDict.items():
            testrunSequence[key] = [sequence["SequenceClass"], {}]
            for field, value in sequence.items():
                testRunAttributes[GC.STRUCTURE_TESTCASESEQUENCE][key][1][field] = value


        xlsTab = self._getTab("TestCase")
        if xlsTab:
            lTestCaseDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            lTestCaseDict = {1: {"TestCaseSequenceNumber": 1,
                             "TestCaseNumber": 1,
                             "TestCaseType": GC.KWARGS_BROWSER,
                             GC.KWARGS_BROWSER: GC.BROWSER_FIREFOX,
                             GC.BROWSER_ATTRIBUTES: ""}}
        for key, testCase in lTestCaseDict.items():
            testSequenceRoot = testrunSequence[testCase["TestCaseSequenceNumber"]][1]
            testSequenceRoot[GC.STRUCTURE_TESTCASE] = {testCase["TestCaseNumber"]: []}
            testSequenceRoot[GC.STRUCTURE_TESTCASE][testCase["TestCaseNumber"]].append(testCase["TestCaseClass"])
            testSequenceRoot[GC.STRUCTURE_TESTCASE][testCase["TestCaseNumber"]].append(
                {"TestCaseType": testCase["TestCaseType"],
                 GC.KWARGS_BROWSER: testCase["Browser"],
                 GC.BROWSER_ATTRIBUTES: testCase["BrowserAttributes"]
                 })

        xlsTab = self._getTab("TestStep")
        if xlsTab:
            lStepDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        else:
            lStepDict = {"TestCaseSequenceNumber": 1,
                         "TestCaseNumber": 1,
                         "TestStepNumber": 1,
                         "TestStepClass": 'baangt.TestSteps.TestStepMaster'}
        for key, testStep in lStepDict.items():
            testStepRoot = testrunSequence[testStep["TestCaseSequenceNumber"]][1]
            testStepRoot = testStepRoot[GC.STRUCTURE_TESTCASE][testStep["TestCaseNumber"]]
            if len(testStepRoot) == 2:
                testStepRoot.append({GC.STRUCTURE_TESTSTEP: {}})

            testStepRoot = testStepRoot[2][GC.STRUCTURE_TESTSTEP]
            testStepRoot[testStep["TestStepNumber"]] = testStep["TestStepClass"]


        xlsTab = self._getTab("TestStepExecution")
        lExecDict = self.getRowsWithHeadersAsDict(xlsTab=xlsTab)
        for key, execLine in lExecDict.items():
            lSequence = self.testRunUtils.getSequenceByNumber(testRunName=self.fileName,
                                                              sequence=execLine.get("TestCaseSequenceNumber",1))
            lTestCase = self.testRunUtils.getTestCaseByNumber(sequence=lSequence,
                                                              testcaseNumber=execLine.get("TestCaseNumber",1))
            lTestStep = self.testRunUtils.getTestStepByNumber(testCase=lTestCase,
                                                              testStepNumber=execLine.get("TestStepNumber",1))
            # if this TestStep is still just a String, then it needs to be converted into a List with Dict
            # to hold the TestexecutionSteps in this Dict.
            if isinstance(lTestStep, str):
                lTestCase[2][GC.STRUCTURE_TESTSTEP][execLine.get("TestStepNumber",1)] = \
                    [lTestStep, {GC.STRUCTURE_TESTSTEPEXECUTION: {}}]
                lTestStep = self.testRunUtils.getTestStepByNumber(testCase=lTestCase,
                                                                  testStepNumber=execLine.get("TestStepNumber",1))

            lTestStep[1][GC.STRUCTURE_TESTSTEPEXECUTION][execLine["TestStepExecutionNumber"]] = {}
            for lkey, value in execLine.items():
                lTestStep[1][GC.STRUCTURE_TESTSTEPEXECUTION][execLine["TestStepExecutionNumber"]][lkey] = value

    def getRowsWithHeadersAsDict(self, xlsTab):
        lRetDict = {}

        for row in range(1, xlsTab.nrows):
            lRetDict[row] = {}
            for col in range(0, xlsTab.ncols):
                lRetDict[row][xlsTab.cell_value(0, col)] = self.replaceFieldValueWithValueOfConstant(
                    xlsTab.cell_value(row, col))
        return lRetDict

    def _getRowAsList(self, xlsTab, colNumber=0):
        l_fields = []
        for row in range(1, xlsTab.nrows):
            l_fields.append(self.replaceFieldValueWithValueOfConstant(xlsTab.cell_value(row, colNumber)))
        return l_fields

    def _getValueFromList(self, xlsTab, searchField):
        # Searches for Value in Column 1 and returns Value from Column 2 if found.
        for row in range(1, xlsTab.nrows):
            if xlsTab.cell_value(row, 1) == searchField:
                return self.replaceFieldValueWithValueOfConstant(xlsTab.cell_value(row, 2))

    def _getTab(self, tabName):
        try:
            worksheet = self.excelFile.sheet_by_name(tabName)
        except Exception as e:
            return None
        return worksheet

    def replaceFieldValueWithValueOfConstant(self, value):
        if isinstance(value, float):
            if value % 1 == 0:
                return int(value)
        elif isinstance(value, int):
            return value

        value = value.strip()
        value = utils.replaceFieldValueWithValueOfConstant(value)

        if value[0:5] == "CGC.":
            value = getattr(globals()[value.split(".")[0]], value.split(".")[1])

        return value



import PySimpleGUI as sg
import glob
import os
import subprocess
import configparser
import baangt.base.GlobalConstants as GC
from baangt.base.utils import utils
import logging
import json

logger = logging.getLogger("pyC")


class UI:
    """
    Provides a simple UI for Testrun-Execution
    """
    def __init__(self):
        # Todo 1: Save last used directory to local config File
        # Todo 2: Use the parameters from Globals and provide an option to change them
        self.configFile = None
        self.configFiles = []
        self.testRunFile = None
        self.testRunFiles = []
        self.configContents = {}
        self.window = None

        self.directory = None
        self.readConfig()
        self.getConfigFilesInDirectory()

        self.startWindow()

    def getLayout(self):
        lLayout = [[sg.Text("Select Directory, Testrun and Global settings to use:")],
                   [sg.Text("Directory", size=(15,1)),
                    sg.In(key="-directory-", size=(30,1), enable_events=True, default_text=self.directory),
                    sg.FolderBrowse(initial_folder=os.getcwd(), enable_events=True)],
                   [sg.Text("TestRun", size=(15,1)),
                    sg.InputCombo(self.testRunFiles, key="testRunFile", default_value=self.testRunFile, size=(30,1))],
                   [sg.Text("Global Settings", size=(15,1)),
                    sg.InputCombo(self.configFiles, key="configFile", default_value=self.configFile, enable_events=True, size=(30,1))]]

        if self.configContents:
            lLayout.append([sg.Text("-"*10 + f"Settings of file {self.configFile}" + "-"*10)])
            for key, value in self.configContents.items():
                lLayout.append([sg.In(key, key="-attrib-" + key, size=(25,1)), sg.In(key="-val-"+key, size=(30,1), default_text=value)])
            for i in range(0,4):
                lLayout.append([sg.In(key=f"-newField-{i}", size=(25,1)), sg.In(key=f"-newValue-{i}", size=(30,1))])
            lLayout.append([sg.Button('Save'), sg.Button("SaveAs"), sg.Button('Exit'), sg.Button("Execute TestRun")])
        else:
            lLayout.append([sg.Button('Exit')])

        return lLayout

    def startWindow(self):
        sg.theme("TanBlue")

        self.window = sg.Window("Baangt interactive Starter", layout=self.getLayout())
        lWindow = self.window
        lWindow.finalize()

        while True:
            lEvent, lValues = lWindow.read()
            if lEvent == "Exit":
                break
            if lValues.get('-directory-') != self.directory:
                self.directory = lValues.get("-directory-")
                self.getConfigFilesInDirectory()
                lWindow['configFile'].update(values=self.configFiles, value="")
                lWindow['testRunFile'].update(values=self.testRunFiles, value="")
                lValues['configFile'] = ""

            if lValues["testRunFile"]:
                self.testRunFile = lValues["testRunFile"]

            if lValues["configFile"]:
                if lValues["configFile"] != self.configFile:
                    self.configFile = lValues['configFile']
                    self.readContentsOfGlobals()
                    lWindow = self.reopenWindow(lWindow)

            if lEvent == 'Save':
                lWindow = self.saveConfigFileProcedure(lWindow, lValues)

            if lEvent == 'SaveAs':
                self.configFile = sg.popup_get_text("New Name of Configfile:")
                if len(self.configFile) > 0:
                    lWindow = self.saveConfigFileProcedure(lWindow, lValues)

            if lEvent == "Execute TestRun":
                self.runTestRun()


        self.saveInteractiveGuiConfig()

    def saveConfigFileProcedure(self, lWindow, lValues):
        # receive updated fields and values to store in JSON-File
        self.modifyValuesOfConfigFileInMemory(lValues)
        self.saveContentsOfConfigFile()
        lWindow = self.reopenWindow(lWindow)
        return lWindow

    def reopenWindow(self, lWindow):
        lWindow.close()
        self.window = sg.Window("Baangt interactive Starter", layout=self.getLayout())
        lWindow = self.window
        lWindow.finalize()
        return lWindow

    def runTestRun(self):
        if not self.configFile:
            sg.popup_cancel("No Config File selected - can't run")
            return
        if not self.testRunFile:
            sg.popup_cancel("No Testrun File selected - can't run")
            return
        runCmd = f"python baangt.py --run='{self.directory}/{self.testRunFile}' " \
                 f"--globals='{self.directory}/{self.configFile}'"
        logger.info(f"Running command: {runCmd}")
        p = subprocess.run(runCmd, shell=True)
        sg.popup_ok("Testrun finished")

    def getConfigFilesInDirectory(self):
        """Reads *.JSON-Files from directory given in self.directory and builds 2 lists (Testrunfiles and ConfiFiles"""
        self.configFiles = []
        self.testRunFiles = []
        lcwd = os.getcwd()
        os.chdir(self.directory)
        fileList = glob.glob("*.json")
        fileList.extend(glob.glob("*.xlsx"))
        for file in fileList:
            if file[0:4].lower() == 'glob':
                self.configFiles.append(file)
            else:
                self.testRunFiles.append(file)
            pass

        self.configFiles = sorted(self.configFiles)
        self.testRunFiles = sorted(self.testRunFiles)

        os.chdir(lcwd)

    def readContentsOfGlobals(self):
        self.configContents = utils.openJson(self.directory + "/" + self.configFile)

    def saveContentsOfConfigFile(self):

        x = sg.popup_ok_cancel(f"Would write this content: {self.configContents} \n to file: {self.configFile} in path {self.directory}")
        print(x)

        if x == 'OK':
            with open(self.directory + "/" + self.configFile, 'w') as outfile:
                json.dump(self.configContents, outfile, indent=4)
            sg.popup_ok(f"Wrote file {self.directory}/{self.configFile}")

    def modifyValuesOfConfigFileInMemory(self, lValues):
        for key, value in lValues.items():
            if '-attrib-' in key:
                # Existing field - update value from value
                lSearchKey = key.replace("-attrib-","")
                if lSearchKey != value:
                    # an existing variable was changed to a new name. Delete the old one:
                    self.configContents.pop(lSearchKey)
                    lSearchKey = value
                    lSearchVal = lValues['-val-'+key.replace("-attrib-","")]
                else:
                    lSearchVal = lValues['-val-'+lSearchKey]
                if len(lSearchKey) > 0:
                    self.configContents[lSearchKey] = lSearchVal
            elif '-newField-' in key:
                # New field needs to be added to memory:
                lSearchKey = value # the new fieldname
                if len(lSearchKey) > 0:
                    lSearchVal = lValues['-newValue-'+key[-1]]
                    self.configContents[lSearchKey] = lSearchVal
            elif '-val-' in key or '-newValue-':
                pass # Values have been used already above
            else:
                logger.critical(f"Program error. Received something with key {key}, value {value} and no "
                                f"idea what to do with it")

    def saveInteractiveGuiConfig(self):
        config = configparser.ConfigParser()
        config["DEFAULT"] = {"path": self.directory}
        with open("baangt.ini", "w") as configFile:
            config.write(configFile)

    def readConfig(self):
        config = configparser.ConfigParser()
        try:
            config.read("baangt.ini")
            self.directory = config["DEFAULT"]['path']
        except Exception as e:
            self.directory = os.getcwd()
            pass





